#!/usr/bin/env python
# Copyright 2025 Pride Leong <lykling.lyk@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""controller
"""

import curses
import threading
import time
from collections import deque

import can
import cantools
from cyber.python.cyber_py3 import cyber
from modules.common_msgs.chassis_msgs import chassis_pb2
from modules.common_msgs.control_msgs import control_cmd_pb2

from akbc import screen


class KeyboardController:

    def __init__(self, scr, dbc_file='./vehicle.dbc', device='can0'):
        self.scr = scr
        self.node = cyber.Node('simple-vehicle-controller')
        self.reader = self.node.create_reader('/apollo/canbus/chassis',
                                              chassis_pb2.Chassis,
                                              self.on_chassis)
        self.writer = self.node.create_writer('/apollo/control',
                                              control_cmd_pb2.ControlCommand)
        self.db = cantools.database.load_file(dbc_file)
        self.canbus = can.Bus(
            interface='socketcan',
            channel=device,
            bitrate=500000,
        )
        self.threads = []
        self.period = 0.1
        self.gear_location_idx = 0
        self.motion_mode_idx = 0

        self.controls = {
            'driving_mode': 0,
            'steering_target': 0.0,
            'steering_rate': 0.0,
            'acceleration': 0.0,
            'gear_location': 0,
            'motion_mode': 1,
        }
        self.chassis = {
            'engine_started': 0.0,
            'speed_mps': 0,
            'gear_location': 0,
            'driving_mode': 0,
            'motion_mode': 0,
            'stickcontrol_event': 0,
            'steering_angle': 0.0,
            'steering_angle_speed': 0.0,
            'longitudinal_acceleration': 0.0,
        }

        self.logs = deque(maxlen=20)
        self.control_cmd_msg = control_cmd_pb2.ControlCommand()

        # register keyboard handlers
        self.scr.on(ord('q'), self.handle_input)
        self.scr.on(curses.KEY_UP, self.handle_input)
        self.scr.on(curses.KEY_DOWN, self.handle_input)
        self.scr.on(curses.KEY_LEFT, self.handle_input)
        self.scr.on(curses.KEY_RIGHT, self.handle_input)
        self.scr.on(ord('m'), self.handle_input)
        self.scr.on(ord('s'), self.handle_input)
        self.scr.on(ord('g'), self.handle_input)

        self._stop_event = threading.Event()

    def log(self, message):
        """log
        """
        self.logs.append(f'{time.strftime("%H:%M:%S")} - {message}')

    def draw_vehicle(self):
        """draw vehicle
        """
        self.scr.vehicle_win.draw_raw([
            r'   _____',
            u'  /_____\\',
            r'  |o   o|',
            r'  |     |',
            r'  |     |',
            r'  |o   o|',
            u'  |-----|',
        ], 2, 2)

    def draw_controls(self):
        """draw controls
        """
        self.scr.controls_win.draw('Controls:', [
            'Up/Down: Acceleration/Brake',
            'Left/Right: Turn Left/Right',
            's: Switch autodriving mode',
            'm: Switch motion mode',
            'g: Switch gear location',
            'Q: Quit',
        ])

    def draw_status(self):
        """draw chassis status
        """
        self.scr.status_win.draw('Chassis:', [
            f'engine_started: {self.chassis["engine_started"]}',
            f'gear_location: {self.chassis["gear_location"]}',
            f'driving_mode: {self.chassis["driving_mode"]}',
            f'steering_angle: {self.chassis["steering_angle"]} deg',
            (r'steering_angle_speed: '
             f'{self.chassis["steering_angle_speed"]} deg/s'),
            f'speed_mps: {self.chassis["speed_mps"]} m/s',
            (r'longitudinal_acceleration: '
             f'{self.chassis["longitudinal_acceleration"]} m/s^2'),
            f'stickcontrol_event: {self.chassis["stickcontrol_event"]}',
            f'motion_mode: {self.chassis["motion_mode"]}',
        ], curses.A_BOLD | curses.color_pair(1))

    def draw_logs(self):
        """draw operation logs
        """
        content = []
        for log in self.logs:
            content.append(log[:self.scr.log_win.width - 4])

        self.scr.log_win.draw('Logs:', content,
                              curses.A_BOLD | curses.color_pair(2))

    def on_update_screen(self):
        """on_update_screen
        """
        self.draw_vehicle()
        self.draw_controls()
        self.draw_status()
        self.draw_logs()

    def handle_input(self, key):
        """handle_input
        """
        if key == curses.KEY_UP:
            if self.controls['acceleration'] >= 0:
                self.controls['acceleration'] += 0.1
            else:
                self.controls['acceleration'] = 0
            self.log(f'accelerated to {self.controls["acceleration"]} m/s^2')

        elif key == curses.KEY_DOWN:
            if self.controls['acceleration'] <= 0:
                self.controls['acceleration'] -= 0.1
            else:
                self.controls['acceleration'] = 0
            self.log(f'braked to {self.controls["acceleration"]} m/s^2')

        elif key == curses.KEY_LEFT:
            if self.controls['motion_mode'] == 2:
                self.controls['steering_rate'] += 0.5
                self.log(
                    f'steering_rate: {self.controls["steering_rate"]} deg/s')
            else:
                self.controls['steering_target'] += 0.5
                self.log(
                    f'steering_target: {self.controls["steering_target"]} deg')

        elif key == curses.KEY_RIGHT:
            if self.controls['motion_mode'] == 2:
                self.controls['steering_rate'] -= 0.5
                self.log(
                    f'steering_rate: {self.controls["steering_rate"]} deg/s')
            else:
                self.controls['steering_target'] -= 0.5
                self.log(
                    f'steering_target: {self.controls["steering_target"]} deg')

        elif key == ord('m'):
            # switch motion mode
            motions = [1, 2, 3, 4, 5]
            self.motion_mode_idx = (self.motion_mode_idx + 1) % len(motions)
            self.controls['motion_mode'] = motions[self.motion_mode_idx]
            self.log(f'set motion_mode to {self.controls["motion_mode"]}.')

        elif key == ord('s'):
            # switch autodriving mode
            if self.controls['driving_mode'] == 0:
                self.controls['driving_mode'] = 1
            else:
                self.controls['driving_mode'] = 0
            self.log(f'driving_mode to {self.controls["driving_mode"]}.')

        elif key == ord('g'):
            # switch gear location
            gears = [0, 1, 2, 3]
            self.gear_location_idx = (self.gear_location_idx + 1) % len(gears)
            self.controls['gear_location'] = gears[self.gear_location_idx]
            self.log(f'set gear_location to {self.controls["gear_location"]}.')

    def on_chassis(self, chassis):
        """on_chassis
        """
        # TODO(All): select fields via config
        self.chassis['engine_started'] = chassis.engine_started
        self.chassis['speed_mps'] = chassis.speed_mps
        self.chassis['gear_location'] = chassis.gear_location
        self.chassis['driving_mode'] = chassis.driving_mode
        self.chassis['motion_mode'] = chassis.motion_mode
        self.chassis['stickcontrol_event'] = chassis.stickcontrol_event
        self.chassis['steering_angle'] = chassis.steering_angle
        self.chassis['steering_angle_speed'] = chassis.steering_angle_speed
        self.chassis[
            'longitudinal_acceleration'] = chassis.longitudinal_acceleration

    def get_control_command(self):
        """get_control_command
        """
        self.control_cmd_msg.Clear()
        self.control_cmd_msg.pad_msg.driving_mode = self.controls[
            'driving_mode']
        if self.controls['driving_mode'] == 1:
            self.control_cmd_msg.pad_msg.action = 1
        else:
            self.control_cmd_msg.pad_msg.action = 0
        self.control_cmd_msg.motion_mode = self.controls['motion_mode']
        self.control_cmd_msg.gear_location = self.controls['gear_location']
        self.control_cmd_msg.acceleration = self.controls['acceleration']
        self.control_cmd_msg.steering_target = self.controls['steering_target']
        self.control_cmd_msg.steering_rate = self.controls['steering_rate']
        return self.control_cmd_msg

    def on_update(self):
        """on_update
        """
        while self.running():
            cmd = self.get_control_command()
            self.writer.write(cmd)

            self.on_update_screen()
            time.sleep(self.period)

    def running(self):
        """running
        """
        return not self._stop_event.is_set()

    def run(self):
        self.log("System started")

        update_thread = threading.Thread(target=self.on_update)
        self.threads.append(update_thread)
        update_thread.start()

        update_thread.join()

    def shutdown(self):
        """shutdown
        """
        self._stop_event.set()


def main(stdscr):
    """main entry
    """
    cyber.init()

    curses.use_default_colors()

    scr = screen.Screen(stdscr)
    controller = KeyboardController(scr)

    threading.Thread(target=scr.run).start()

    try:
        controller.run()
    except KeyboardInterrupt:
        controller.shutdown()
        controller.log("System interrupted by user")
    finally:
        controller.shutdown()
        cyber.shutdown()


if __name__ == '__main__':
    curses.wrapper(main)



#!/usr/bin/env python
"""simple vechile
"""
import curses
import math
import signal
import threading
import time
import traceback

import can
import cantools
import cantools.database
import cantools.database.can.database
import numpy as np
from cyber.python.cyber_py3 import cyber
from modules.common_msgs.localization_msgs import localization_pb2
from scipy.spatial.transform import Rotation

from akbc import screen

exit_event = threading.Event()


def clamp(value, min_value, max_value):
    """clamp
    """
    if value < min_value:
        return min_value
    elif value > max_value:
        return max_value
    else:
        return value


class SimpleVehicle:

    class Attrs:
        """Attrs
        """

        def __init__(self, vin='12345678901234567', wheelbase=0.96):
            """__init__
            """
            self.vin = vin
            self.L = wheelbase

    class State:
        """State
        """

        def __init__(self,
                     initial_x=0.0,
                     initial_y=0.0,
                     initial_yaw=0.0,
                     wheelbase=0.96,
                     vin='12345678901234567'):
            """__init__
            """
            self.auto_driving = 0
            self.takeover = 0
            # pose
            self.x = initial_x
            self.y = initial_y
            self.z = 0.0
            # attitude
            self.pitch = 0.0
            self.roll = 0.0
            self.yaw = initial_yaw

            # displacement
            self.s = 0.0

            # velocity in vehicle axis
            self.velocity_x = 0.0
            self.velocity_y = 0.0
            self.velocity_z = 0.0
            # acceleration in vehicle axis
            self.acceleration_x = 0.0
            self.acceleration_y = 0.0
            self.acceleration_z = 0.0

            # angular velocity
            self.angular_rate_pitch = 0.0
            self.angular_rate_roll = 0.0
            self.angular_rate_yaw = 0.0

            # steering
            self.steering = 0.0
            self.steering_rate = 0.0

            # motion
            self.motion_mode = 1

            # gear
            self.gear = 0

            # error state
            self.errno = 0

    class Messages:
        """Messages
        """

        def __init__(self, database):
            """__init__
            """
            self.db = database

    def __init__(
        self,
        scr,
        dbc_file,
        device,
        initial_x=0.0,
        initial_y=0.0,
        initial_yaw=0.0,
        vin='12345678901234567',
        wheelbase=0.96,
    ):
        """__init__
        """
        self._init_args = {
            'dbc_file': dbc_file,
            'device': device,
            'initial_x': initial_x,
            'initial_y': initial_y,
            'initial_yaw': initial_yaw,
            'vin': vin,
            'wheelbase': wheelbase,
        }
        self.scr = scr
        self.db = cantools.database.load_file(dbc_file)
        if not isinstance(self.db, cantools.database.can.database.Database):
            raise TypeError('dbc_file should be a can database file')

        self.can_bus = can.Bus(
            interface='socketcan',
            channel=device,
            bitrate=500000,
        )

        self.attrs = self.Attrs(vin, wheelbase)
        self.state = self.State(initial_x, initial_y, initial_yaw)

        self.control_messages = [
            self.db.get_message_by_name('ADAS_Heartbeat_Command'),
            self.db.get_message_by_name('ADAS_Vehicle_Mode_Command'),
            self.db.get_message_by_name('ADAS_Vehicle_Drive_Command'),
            self.db.get_message_by_name('IMU_Info'),
        ]
        self.report_messages = [
            self.db.get_message_by_name('VCU_Vehicle_Mode_Report'),
            self.db.get_message_by_name('VCU_Vehicle_Drive_Report'),
            self.db.get_message_by_name('VCU_Error_Report'),
            self.db.get_message_by_name('VCU_Vehicle_VIN_Report_01'),
            self.db.get_message_by_name('VCU_Vehicle_VIN_Report_02'),
            self.db.get_message_by_name('VCU_Vehicle_VIN_Report_03'),
            self.db.get_message_by_name('StickControl'),
        ]
        self.report_messages_last_send_time = {
            self.db.get_message_by_name('VCU_Vehicle_Mode_Report'): 0.0,
            self.db.get_message_by_name('VCU_Vehicle_Drive_Report'): 0.0,
            self.db.get_message_by_name('VCU_Error_Report'): 0.0,
            self.db.get_message_by_name('VCU_Vehicle_VIN_Report_01'): 0.0,
            self.db.get_message_by_name('VCU_Vehicle_VIN_Report_02'): 0.0,
            self.db.get_message_by_name('VCU_Vehicle_VIN_Report_03'): 0.0,
            self.db.get_message_by_name('StickControl'): 0.0,
        }

        self.bind_report_messages()

        self.node = cyber.Node('simple-vehicle')
        self.pose_writer = self.node.create_writer(
            '/apollo/localization/pose', localization_pb2.LocalizationEstimate)

        self.paused = False
        self._stop_event = threading.Event()

    def bind_report_messages(self):
        """bind_report_messages
        """
        self.report_getters = {}
        # VCU_Vehicle_Drive_Report
        vcu_vehicle_mode_report = self.db.get_message_by_name(
            'VCU_Vehicle_Mode_Report')
        self.report_getters[vcu_vehicle_mode_report.get_signal_by_name(
            'AutoDriving_STATE')] = lambda: self.state.auto_driving
        self.report_getters[vcu_vehicle_mode_report.get_signal_by_name(
            'DrivingMode')] = lambda: self.state.motion_mode
        self.report_getters[vcu_vehicle_mode_report.get_signal_by_name(
            'Gear')] = lambda: self.state.gear
        # VCU_Vehicle_Drive_Report
        vcu_vehicle_mode_report = self.db.get_message_by_name(
            'VCU_Vehicle_Drive_Report')
        self.report_getters[vcu_vehicle_mode_report.get_signal_by_name(
            'SteerAngle')] = lambda: self.state.steering
        self.report_getters[vcu_vehicle_mode_report.get_signal_by_name(
            'VehicleAngleSpeed')] = lambda: math.degrees(self.state.
                                                         angular_rate_yaw)
        self.report_getters[vcu_vehicle_mode_report.get_signal_by_name(
            'LongitudinalAcceleration')] = lambda: self.state.acceleration_x
        self.report_getters[vcu_vehicle_mode_report.get_signal_by_name(
            'VehicleSpeed')] = lambda: math.fabs(self.state.velocity_x)
        # VCU_Error_Report
        vcu_error_report = self.db.get_message_by_name('VCU_Error_Report')
        self.report_getters[vcu_error_report.get_signal_by_name(
            'ErrorCode')] = lambda: self.state.errno
        # VCU_Vehicle_VIN_Report_01
        vcu_vehicle_vin_report_01 = self.db.get_message_by_name(
            'VCU_Vehicle_VIN_Report_01')
        self.report_getters[vcu_vehicle_vin_report_01.get_signal_by_name(
            'VIN00')] = lambda: ord(self.attrs.vin[0])
        self.report_getters[vcu_vehicle_vin_report_01.get_signal_by_name(
            'VIN01')] = lambda: ord(self.attrs.vin[1])
        self.report_getters[vcu_vehicle_vin_report_01.get_signal_by_name(
            'VIN02')] = lambda: ord(self.attrs.vin[2])
        self.report_getters[vcu_vehicle_vin_report_01.get_signal_by_name(
            'VIN03')] = lambda: ord(self.attrs.vin[3])
        self.report_getters[vcu_vehicle_vin_report_01.get_signal_by_name(
            'VIN04')] = lambda: ord(self.attrs.vin[4])
        self.report_getters[vcu_vehicle_vin_report_01.get_signal_by_name(
            'VIN05')] = lambda: ord(self.attrs.vin[5])
        self.report_getters[vcu_vehicle_vin_report_01.get_signal_by_name(
            'VIN06')] = lambda: ord(self.attrs.vin[6])
        self.report_getters[vcu_vehicle_vin_report_01.get_signal_by_name(
            'VIN07')] = lambda: ord(self.attrs.vin[7])
        # VCU_Vehicle_VIN_Report_02
        vcu_vehicle_vin_report_02 = self.db.get_message_by_name(
            'VCU_Vehicle_VIN_Report_02')
        self.report_getters[vcu_vehicle_vin_report_02.get_signal_by_name(
            'VIN08')] = lambda: ord(self.attrs.vin[8])
        self.report_getters[vcu_vehicle_vin_report_02.get_signal_by_name(
            'VIN09')] = lambda: ord(self.attrs.vin[9])
        self.report_getters[vcu_vehicle_vin_report_02.get_signal_by_name(
            'VIN10')] = lambda: ord(self.attrs.vin[10])
        self.report_getters[vcu_vehicle_vin_report_02.get_signal_by_name(
            'VIN11')] = lambda: ord(self.attrs.vin[11])
        self.report_getters[vcu_vehicle_vin_report_02.get_signal_by_name(
            'VIN12')] = lambda: ord(self.attrs.vin[12])
        self.report_getters[vcu_vehicle_vin_report_02.get_signal_by_name(
            'VIN13')] = lambda: ord(self.attrs.vin[13])
        self.report_getters[vcu_vehicle_vin_report_02.get_signal_by_name(
            'VIN14')] = lambda: ord(self.attrs.vin[14])
        self.report_getters[vcu_vehicle_vin_report_02.get_signal_by_name(
            'VIN15')] = lambda: ord(self.attrs.vin[15])
        # VCU_Vehicle_VIN_Report_03
        vcu_vehicle_vin_report_03 = self.db.get_message_by_name(
            'VCU_Vehicle_VIN_Report_03')
        self.report_getters[vcu_vehicle_vin_report_03.get_signal_by_name(
            'VIN16')] = lambda: ord(self.attrs.vin[16])
        # StickControl
        stickcontrol = self.db.get_message_by_name('StickControl')
        self.report_getters[stickcontrol.get_signal_by_name(
            'TakeOver_CTRL')] = lambda: self.state.takeover

    def on_report(self):
        """on_report
        """
        while self.running():
            # now in millisecond
            now = time.time() * 1000
            for message in self.report_messages:
                if message.cycle_time is None:
                    # not periodic message
                    continue
                if now - self.report_messages_last_send_time[message] < float(
                        message.cycle_time):
                    continue
                data = {}
                for sig in message.signals:
                    if sig in self.report_getters:
                        value = self.report_getters[sig]()
                        if value is not None:
                            data[sig.name] = value
                buffer = message.encode(data)
                msg = can.Message(arbitration_id=message.frame_id,
                                  data=buffer,
                                  is_extended_id=False)
                self.can_bus.send(msg)
                self.report_messages_last_send_time[message] = now
            time.sleep(0.001)

    def on_control_auto_driving(self, auto_driving):
        """on_control_auto_driving
        """
        if auto_driving == self.state.auto_driving:
            return
        self.state.auto_driving = auto_driving
        if auto_driving == 0:
            self.state.velocity_x = 0.0
            self.state.acceleration_x = 0.0
            self.state.steering = 0.0
            self.state.steering_rate = 0.0

    def on_control_motion_mode(self, motion_mode):
        """on_control_motion_mode
        """
        if motion_mode == self.state.motion_mode:
            return
        self.state.motion_mode = motion_mode
        self.state.velocity_x = 0.0
        self.state.acceleration_x = 0.0
        self.state.steering = 0.0
        self.state.steering_rate = 0.0

    def on_control_gear(self, gear):
        """on_control_gear
        """
        if gear == self.state.gear:
            return
        self.state.gear = gear
        self.state.velocity_x = 0.0
        self.state.acceleration_x = 0.0
        self.state.steering = 0.0
        self.state.steering_rate = 0.0

    def on_control_acceleration(self, acceleration, deceleration):
        """on_control_acceleration
        """
        if deceleration > 0:
            self.state.acceleration_x = -deceleration
        else:
            self.state.acceleration_x = acceleration

        self.state.acceleration_x = clamp(self.state.acceleration_x, -10.23,
                                          10.24)

    def on_control_steering(self, steering):
        """on_control_steering
        """
        if self.state.motion_mode == 2:
            return
        self.state.steering = clamp(steering, -163.83, 163.84)

    def on_control_steering_rate(self, steering_rate):
        """on_control_steering_rate
        """
        if self.state.motion_mode != 2:
            return
        self.state.steering_rate = clamp(steering_rate, -327.68, 327.67)
        r = self.attrs.L / 2.0
        self.state.velocity_x = self.state.steering_rate * math.pi / 180 * r

    def on_control(self):
        """on_control
        """
        while self.running():
            message = self.can_bus.recv(timeout=1)
            if message is None:
                continue
            # print('recv message: ', hex(message.arbitration_id),
            #       message.data.hex())
            if message.arbitration_id == self.db.get_message_by_name(
                    'ADAS_Heartbeat_Command').frame_id:
                pass
            elif message.arbitration_id == self.db.get_message_by_name(
                    'ADAS_Vehicle_Mode_Command').frame_id:
                data = self.db.get_message_by_name(
                    'ADAS_Vehicle_Mode_Command').decode(message.data)
                # AutoDriving_CTRL
                self.on_control_auto_driving(data['AutoDriving_CTRL'].value)
                # DrivingMode
                self.on_control_motion_mode(data['DrivingMode_Target'].value)
                # Gear
                self.on_control_gear(data['Gear_Target'].value)
            elif message.arbitration_id == self.db.get_message_by_name(
                    'ADAS_Vehicle_Drive_Command').frame_id:
                data = self.db.get_message_by_name(
                    'ADAS_Vehicle_Drive_Command').decode(message.data)
                # SteerAngle
                self.on_control_steering(data['SteerAngle_Target'])
                # VehicleAngleSpeed
                self.on_control_steering_rate(data['VehicleAngleSpeed_Target'])
                self.on_control_acceleration(
                    data['LongitudinalAcceleration_Target'],
                    data['LongitudinalDeceleration_Target'])
            elif message.arbitration_id == self.db.get_message_by_name(
                    'IMU_Info').frame_id:
                pass
            # time.sleep(0.1)

    def on_update_ackermann(self, dt):
        """on_update_ackermann
        """
        pass
        v = math.fabs(self.state.velocity_x)
        v = clamp(v + self.state.acceleration_x * dt, 0.0, 20.47)
        if self.state.gear == 2:
            self.state.velocity_x = -v
        else:
            self.state.velocity_x = v
        if v < 1e-3:
            v = 0
            return

        heading = self.state.yaw
        steering = self.state.steering
        wheelbase = self.attrs.L

        tan_delta = math.tan(steering * math.pi / 180)
        omega = 0.0
        if math.fabs(tan_delta) > 1e-6:
            omega = v * tan_delta / wheelbase
            if self.state.motion_mode == 5:
                omega = omega * 2.0
            omega = clamp(omega, -327.68 * math.pi / 180,
                          327.67 * math.pi / 180)
        heading = math.fmod(heading + omega * dt, 2 * math.pi)
        ds = math.fabs(self.state.velocity_x) * dt
        dx = ds * math.cos((self.state.yaw + heading) / 2)
        dy = ds * math.sin((self.state.yaw + heading) / 2)
        self.state.x += dx
        self.state.y += dy
        self.state.s += ds
        self.state.yaw = heading
        self.state.angular_rate_yaw = omega

    def on_update_spot_turn(self, dt):
        """on_update_spot_turn
        """
        angular_delta = self.state.steering_rate * dt * math.pi / 180
        self.state.angular_rate_yaw = clamp(angular_delta / dt,
                                            -327.68 * math.pi / 180,
                                            327.67 * math.pi / 180)
        self.state.velocity_x = self.state.angular_rate_yaw * self.attrs.L / 2
        v = math.fabs(self.state.velocity_x)
        if v < 1e-3:
            v = 0
            self.angular_rate_yaw = 0
            return

        direction = 1.0
        if self.state.gear == 4:
            direction = 1.0
        else:
            direction = -1.0
        heading = math.fmod(self.state.yaw + direction * angular_delta,
                            2 * math.pi)
        r = self.attrs.L / 2
        self.state.x = self.state.x + r * (math.cos(self.state.yaw) -
                                           math.cos(heading))
        self.state.y = self.state.y + r * (math.sin(self.state.yaw) -
                                           math.sin(heading))
        self.state.yaw = heading

        ds = v * dt
        self.state.s += ds

    def on_update_sideway(self, dt):
        """on_update_sideway
        """
        v = math.fabs(self.state.velocity_x)
        v = clamp(v + self.state.acceleration_x * dt, 0.0, 20.47)
        if self.state.gear == 2:
            self.state.velocity_x = -v
        else:
            self.state.velocity_x = v
        heading = self.state.yaw
        dx = self.state.velocity_x * math.cos(heading + math.pi / 2) * dt
        dy = self.state.velocity_x * math.sin(heading + math.pi / 2) * dt
        self.state.x += dx
        self.state.y += dy
        ds = math.fabs(self.state.velocity_x) * dt
        self.state.s += ds

    def on_update_crabwalk(self, dt):
        """on_update_crabwalk
        """
        v = math.fabs(self.state.velocity_x)
        v = clamp(v + self.state.acceleration_x * dt, 0.0, 20.47)
        if self.state.gear == 2:
            self.state.velocity_x = -v
        else:
            self.state.velocity_x = v
        heading = self.state.yaw
        angular = math.fmod(heading + self.state.steering * math.pi / 180,
                            math.pi * 2)
        dx = self.state.velocity_x * math.cos(angular) * dt
        dy = self.state.velocity_x * math.sin(angular) * dt
        ds = math.fabs(self.state.velocity_x) * dt
        self.state.x += dx
        self.state.y += dy
        self.state.s += ds

    def on_update(self):
        """update_vehicle_attrs
        """
        dt = 0.001
        ticks = 0
        while self.running():
            ticks += 1

            if self.state.auto_driving == 1 and not self.paused:
                if ((self.state.motion_mode == 1
                     or self.state.motion_mode == 5)
                        and (self.state.gear == 4 or self.state.gear == 2)):
                    self.on_update_ackermann(dt)

                elif (self.state.motion_mode == 2
                      and (self.state.gear == 4 or self.state.gear == 2)):
                    self.on_update_spot_turn(dt)

                elif (self.state.motion_mode == 4
                      and (self.state.gear == 4 or self.state.gear == 2)):
                    self.on_update_sideway(dt)

                elif (self.state.motion_mode == 3
                      and (self.state.gear == 4 or self.state.gear == 2)):
                    self.on_update_crabwalk(dt)

            self.scr.world_win.draw('State', [
                f'ticks: {ticks} paused: {self.paused}',
                f'x: {self.state.x:.6f}',
                f'y: {self.state.y:.6f}',
                f'yaw: {self.state.yaw:.6f}'
                f' {math.degrees(self.state.yaw):.6f}',
                f'omega: {self.state.angular_rate_yaw:.6f}',
                f'vx: {self.state.velocity_x:.6f}',
                f'ax: {self.state.acceleration_x:.6f}',
                f's: {self.state.s:.6f}',
                f'steering: {self.state.steering}',
                f'steering_rate: {self.state.steering_rate}',
            ], curses.A_BOLD | curses.color_pair(1))

            time.sleep(dt)

    def mrf_to_vrf(self, orientation, mrf):
        """mrf_to_vrf
        """
        # mrf: vehicle axis
        # vrf: world axis
        # orientation: quaternion
        rotation = Rotation.from_quat(orientation)
        vrf = np.mat(rotation.inv().as_matrix()) * np.mat(mrf).T
        return [
            vrf.getA1()[0],
            vrf.getA1()[1],
            vrf.getA1()[2],
        ]

    def on_pose(self):
        """on_pose
        """
        seq = 0
        while self.running():
            seq += 1
            pose_msg = localization_pb2.LocalizationEstimate()
            pose_msg.Clear()
            pose_msg.header.timestamp_sec = time.time()
            pose_msg.header.module_name = 'SimpleVehicle'
            pose_msg.header.sequence_num = seq
            pose_msg.pose.position.x = self.state.x
            pose_msg.pose.position.y = self.state.y
            pose_msg.pose.position.z = self.state.z
            orientation = Rotation.from_euler(
                'xyz', [0, 0, self.state.yaw - math.pi / 2]).as_quat()
            pose_msg.pose.orientation.qx = orientation[0]
            pose_msg.pose.orientation.qy = orientation[1]
            pose_msg.pose.orientation.qz = orientation[2]
            pose_msg.pose.orientation.qw = orientation[3]
            pose_msg.pose.heading = self.state.yaw
            linear_velocity = [
                self.state.velocity_x * math.cos(self.state.yaw),
                self.state.velocity_x * math.sin(self.state.yaw),
                self.state.velocity_z,
            ]
            pose_msg.pose.linear_velocity.x = linear_velocity[0]
            pose_msg.pose.linear_velocity.y = linear_velocity[1]
            pose_msg.pose.linear_velocity.z = linear_velocity[2]
            linear_accel = [
                self.state.acceleration_x * math.cos(self.state.yaw),
                self.state.acceleration_x * math.sin(self.state.yaw),
                self.state.acceleration_z,
            ]
            pose_msg.pose.linear_acceleration.x = linear_accel[0]
            pose_msg.pose.linear_acceleration.y = linear_accel[1]
            pose_msg.pose.linear_acceleration.z = linear_accel[2]
            angular_velocity = [
                self.state.angular_rate_pitch,
                self.state.angular_rate_roll,
                self.state.angular_rate_yaw,
            ]
            pose_msg.pose.angular_velocity.x = angular_velocity[0]
            pose_msg.pose.angular_velocity.y = angular_velocity[1]
            pose_msg.pose.angular_velocity.z = angular_velocity[2]

            linear_accel_vrf = self.mrf_to_vrf(orientation, linear_accel)
            pose_msg.pose.linear_acceleration_vrf.x = linear_accel_vrf[0]
            pose_msg.pose.linear_acceleration_vrf.y = linear_accel_vrf[1]
            pose_msg.pose.linear_acceleration_vrf.z = linear_accel_vrf[2]

            angular_velocity_vrf = self.mrf_to_vrf(orientation,
                                                   angular_velocity)
            pose_msg.pose.angular_velocity_vrf.x = angular_velocity_vrf[0]
            pose_msg.pose.angular_velocity_vrf.y = angular_velocity_vrf[1]
            pose_msg.pose.angular_velocity_vrf.z = angular_velocity_vrf[2]

            self.pose_writer.write(pose_msg)
            time.sleep(0.01)

    def pause(self):
        """pause
        """
        self.paused = True

    def resume(self):
        """resume
        """
        self.paused = False

    def reset(self):
        """reset
        """
        self.state.x = self._init_args['initial_x']
        self.state.y = self._init_args['initial_y']
        self.state.yaw = self._init_args['initial_yaw']
        self.state.s = 0.0
        self.state.velocity_x = 0.0
        self.state.acceleration_x = 0.0
        self.state.angular_rate_yaw = 0.0
        self.state.steering = 0.0
        self.state.steering_rate = 0.0
        self.state.motion_mode = 1
        self.state.gear = 0

    def running(self):
        """running
        """
        return not self._stop_event.is_set()

    def run(self):
        """run
        """

        self.control = threading.Thread(target=self.on_control, )
        self.report = threading.Thread(target=self.on_report, )
        self.update = threading.Thread(target=self.on_update, )
        self.pose = threading.Thread(target=self.on_pose, )

        self.control.start()
        self.report.start()
        self.update.start()
        self.pose.start()

        self.control.join()
        self.report.join()
        self.update.join()
        self.pose.join()

    def shutdown(self):
        """shutdown
        """
        self._stop_event.set()


def main():
    """main entry
    """
    cyber.init()
    stdscr = curses.initscr()
    scr = screen.Screen(stdscr)
    vehicle = SimpleVehicle(scr, './vehicle.dbc', 'can0')

    def signal_handler(sig, frame):
        """signal_handler
        """
        print('exiting...')
        exit_signals = [signal.SIGINT, signal.SIGTERM]
        if sig in exit_signals:
            exit_event.set()
            vehicle.shutdown()
            cyber.shutdown()
        else:
            # unhandled signal
            print(f'Unhandled signal: {sig}', traceback.print_stack(frame))

    signal.signal(signal.SIGINT, signal_handler)

    try:
        vehicle.run()
    except KeyboardInterrupt:
        exit_event.set()
        vehicle.shutdown()
        cyber.shutdown()
    finally:
        exit_event.set()
        vehicle.shutdown()
        cyber.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""canbus monitor
"""
import threading
import time

import can
import cantools
import cantools.database
import cantools.database.can.database


class CanbusMonitor:
    """CanbusMonitor
    """

    def __init__(self, dbc_file, device):
        """__init__
        """
        self.db = cantools.database.load_file(dbc_file)
        self.canbus = can.Bus(
            interface='socketcan',
            channel=device,
            bitrate=500000,
        )

        self.canbus_messages = {}
        self.canbus_message_data = {}
        self.canbus_counter = {}
        self.canbus_last = {}
        self.canbus_latency = {}

        self.threads = []
        self._stop_event = threading.Event()

    def on_canbus(self):
        """on_canbus
        """
        while self.running():
            message = self.canbus.recv(1)
            if message is None:
                continue
            available_messages = list(
                map(lambda x: x.frame_id, self.db.messages))
            if message.arbitration_id in available_messages:
                if message.arbitration_id not in self.canbus_counter:
                    self.canbus_counter[message.arbitration_id] = 0
                self.canbus_counter[message.arbitration_id] += 1
                if message.arbitration_id not in self.canbus_last:
                    self.canbus_last[message.arbitration_id] = 0
                last = self.canbus_last[message.arbitration_id]
                now = time.time()
                self.canbus_last[message.arbitration_id] = now
                latency = (now - last) * 1000
                self.canbus_latency[message.arbitration_id] = latency
                spec = self.db.get_message_by_frame_id(message.arbitration_id)
                data = spec.decode(message.data)
                self.canbus_messages[message.arbitration_id] = message
                self.canbus_message_data[message.arbitration_id] = data

    def get_showing_text(self):
        """get_showing_text
        """
        content = []
        for message in self.canbus_messages.values():
            if not message:
                continue
            spec = self.db.get_message_by_frame_id(message.arbitration_id)
            if not spec:
                continue
            content.append(f'{hex(spec.frame_id)}: '
                           f'{spec.length} '
                           f'{self.canbus_counter[spec.frame_id]:6} '
                           f'{self.canbus_latency[spec.frame_id]:8.3f} '
                           f'{message.data.hex()} {spec.name}')
            content.append(f'    {self.canbus_message_data[spec.frame_id]}')
        return content

    def running(self):
        """running
        """
        return not self._stop_event.is_set()

    def run(self):
        canbus_thread = threading.Thread(target=self.on_canbus)
        self.threads.append(canbus_thread)
        canbus_thread.start()

        canbus_thread.join()

    def shutdown(self):
        """shutdown
        """
        self._stop_event.set()



#!/usr/bin/env python
"""apollo keyboard controller
"""
import curses
import signal
import threading
import traceback

import click
from cyber.python.cyber_py3 import cyber

from akbc import canbus_monitor, keyboard_controller, screen, simple_vehicle


@click.command()
@click.option('--disable_keyboard_controller', is_flag=True, default=False)
@click.option('--enable_virtual_vehicle', is_flag=True, default=False)
@click.option('--dbc_file', default='./vehicle.dbc', type=str, help='dbc file')
@click.option('--device', default='can0', type=str, help='can device')
@click.option(
    '--initial_x',
    # default=587362.114900,
    default=586389.25,
    type=float,
    help='initial x')
@click.option(
    '--initial_y',
    # default=4140841.386400,
    default=4140674.01,
    type=float,
    help='initial y')
@click.option(
    '--initial_yaw',
    # default=-0.614823,
    default=2.891656162,
    type=float,
    help='initial yaw')
@click.option(
    '--wheelbase',
    default=0.96,
    type=float,
    help='vehicle wheelbase, default 0.96, unit: m',
)
@click.option(
    '--vin',
    default='12345678901234567',
    type=str,
    help='vehicle vin, default 12345678901234567',
)
def main(**kwargs):
    """main entry
    """
    stdscr = curses.initscr()
    curses.noecho()
    curses.cbreak()
    stdscr.keypad(True)
    scr = screen.Screen(stdscr)

    cyber.init()

    controller = None
    if not kwargs['disable_keyboard_controller']:
        controller = keyboard_controller.KeyboardController(
            scr, kwargs['dbc_file'], kwargs['device'])
    vehicle = None
    if kwargs['enable_virtual_vehicle']:
        vehicle = simple_vehicle.SimpleVehicle(
            scr, kwargs['dbc_file'], kwargs['device'], kwargs['initial_x'],
            kwargs['initial_y'], kwargs['initial_yaw'], kwargs['vin'],
            kwargs['wheelbase'])
    monitor = canbus_monitor.CanbusMonitor(kwargs['dbc_file'],
                                           kwargs['device'])

    def _show_canbus_data():
        """_show_canbus_data
        """
        content = monitor.get_showing_text()
        scr.canbus_win.draw('Canbus:', content,
                            curses.A_BOLD | curses.color_pair(1))

    scr.ontick(_show_canbus_data)

    def _vehicle_pause_handler(_):
        """_vehicle_pause_handler
        """
        if vehicle:
            if vehicle.paused:
                vehicle.resume()
            else:
                vehicle.pause()

    def _vehicle_reset_handler(_):
        """_vehicle_reset_handler
        """
        if vehicle:
            vehicle.reset()

    if vehicle:
        scr.on(ord(' '), _vehicle_pause_handler)
        scr.on(ord('r'), _vehicle_reset_handler)

    threads = []
    screen_thread = threading.Thread(target=scr.run, name='screen')
    threads.append(screen_thread)
    monitor_thread = threading.Thread(target=monitor.run, name='monitor')
    threads.append(monitor_thread)
    if controller:
        controller_thread = threading.Thread(target=controller.run,
                                             name='controller')
        threads.append(controller_thread)
    if vehicle:
        vehicle_thread = threading.Thread(target=vehicle.run, name='vehicle')
        threads.append(vehicle_thread)

    def _exit_screen(_):
        """_exit_curses
        """
        if controller:
            controller.shutdown()
        if vehicle:
            vehicle.shutdown()
        cyber.shutdown()
        monitor.shutdown()
        scr.shutdown()
        stdscr.keypad(False)
        curses.echo()
        curses.endwin()

    scr.on(ord('q'), _exit_screen)

    def _signal_handler(sig, frame):
        """signal_handler
        """
        print('exiting...', sig)
        exit_signals = [signal.SIGINT, signal.SIGTERM]
        if sig in exit_signals:
            if controller:
                controller.shutdown()
            if vehicle:
                vehicle.shutdown()
            cyber.shutdown()
            monitor.shutdown()
            scr.shutdown()
            stdscr.keypad(False)
            curses.echo()
            curses.endwin()
        else:
            # unhandled signal
            print(f'Unhandled signal: {sig}', traceback.print_stack(frame))

    # register signal handlers
    signal.signal(signal.SIGINT, _signal_handler)

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()
    # screen_thread.join()


if __name__ == '__main__':
    main()
